# -*- coding: utf-8 -*-

import pandas as pd
from numpy import NaN, where
from goerr import err
from .db import Db
from .charts import Plot
from .data import Df
from .report import Report
from .errors import Errors
from .messages import Messages
from _ast import arg

__version__ = "0.3"


class DataSwim(Db, Df, Plot, Report, Errors, Messages):
    """
    Main class
    """

    def __init__(self, df=None, db=None):
        """
        Initialize with an empty dataframe
        """
        global __version__
        self.version = __version__
        self.df = df
        self.db = db
        self.x_field = None
        self.y_field = None
        self.chart_obj = None
        self.chart_opts = dict(width=940)
        self.chart_style = dict(color="blue")
        self.engine = "bokeh"
        self.label = None
        self.reports = []
        self.report_path = None
        self.backup_df = None
        self.autoprint = True
        self.errors_handling = "exceptions"

    def __repr__(self):
        num = 0
        if self.df is not None:
            num = len(self.df.index)
        msg = "<DataSwim object | " + str(num) + " rows>"
        return msg

    def new_(self, df=None, db=None, quiet=False):
        """
        Returns a new DataSwim instance from a dataframe
        """
        try:
            ds2 = DataSwim(df, db)
        except Exception as e:
            self.err(e, self.new_, "Can not set new instance")
        if self.autoprint is True and quiet is False:
            self.ok("A new instance was created")
        return ds2


ds = DataSwim()



from gencharts import ChartsGenerator


class Altair():
    """
    A class to handle charts with the Altair library
    """

    def __init__(self, df=None, db=None):
        """
        Initialize
        """
        self.df = df
        self.x_field = None
        self.y_field = None
        self.chart_obj = None
        self.chart_opts = dict(width=940)
        self.chart_style = None
        self.label = None
        self.engine = "bokeh"

    def _get_altair_chart(self, xfield, yfield, chart_type, label, opts, style):
        """
        Get a chart object
        """
        chart = ChartsGenerator()
        # params
        opts["xfield"] = xfield
        opts["yfield"] = yfield
        opts["dataobj"] = self.df
        opts["chart_type"] = chart_type
        # generate
        try:
            chart.serialize(**opts)
            chartobj = chart.serialize(**opts)
        except Exception as e:
            self.err(e, self._get_altair_chart, "Can not get chart object")
            return
        return chartobj



# -*- coding: utf-8 -*-

import os
import holoviews as hv
from bokeh.embed import components
from goerr import err


renderer = hv.renderer('bokeh')


class Report():
    """
    Class to handle reporting
    """

    def stack(self, slug, title, chart_obj=None):
        """
        Get the html for a chart and store it
        """
        script, html = self.get_html(chart_obj)
        title = "<h3>" + title + "</h3>"
        report = dict(slug=slug, title=title, html=script + html)
        self.reports.append(report)

    def file(self, slug, folderpath=None, p=True):
        """
        Writes the html report to a file from the report stack
        """
        if folderpath is None:
            folderpath = self.report_path
        else:
            self.report_path = folderpath
        html = self._header()
        for report in self.reports:
            html += report["title"] + report["html"]
        html += self._footer()
        self._write_file(slug, folderpath, html, p)
        self.reports = []

    def files(self, folderpath=None, p=True):
        """
        Writes the html report to one file per report
        """
        if folderpath is None:
            folderpath = self.report_path
        else:
            self.report_path = folderpath
        for report in self.reports:
            html = report["title"] + report["html"]
            self._write_file(report["slug"], folderpath, html, p)
        self.reports = []

    def get_html(self, chart_obj=None):
        """
        Get the html and script tag for a chart
        """
        if chart_obj is None:
            chart_obj = self.chart_obj
        p = renderer.get_plot(chart_obj).state
        script, div = components(p)
        return script, div

    def _write_file(self, slug, folderpath, html, p=True):
        """
        Writes a chart's html to a file
        """
        # check directories
        if not os.path.isdir(folderpath):
            try:
                os.makedirs(folderpath)
            except Exception as e:
                err.new(e)
        # construct file path
        filepath = folderpath + "/" + slug + ".html"
        #~ write the file
        try:
            filex = open(filepath, "w")
            filex.write(html)
            filex.close()
            if p is True:
                print("File written to", filepath)
        except Exception as e:
            err.new(e)

    def _header(self):
        return """ 
        <!DOCTYPE html>
        <html lang="en">
        <head>
            <meta charset="utf-8">
            <title>Test</title>   
            <link rel="stylesheet" href="https://cdn.pydata.org/bokeh/release/bokeh-0.12.9.min.css" type="text/css" />           
            <script type="text/javascript" src="https://cdn.pydata.org/bokeh/release/bokeh-0.12.9.min.js"></script>
            <script type="text/javascript">
                Bokeh.set_log_level("info");
            </script>
        </head>
        <body>
        """

    def _footer(self):
        return "</body>\n</html>"



from goerr.colors import colors


class Messages():
    """
    A class to handle output
    """

    def ok(self, *msg):
        """
        Prints a message with an ok prefix
        """
        label = colors.green("OK")
        self._msg(label, *msg)

    def info(self, *msg):
        """
        Prints a message with an info prefix
        """
        label = colors.blue("INFO")
        endmsg = self.msg_(label, *msg)
        print(endmsg)

    def warning(self, *msg):
        """
        Prints a warning
        """
        label = colors.yellow("WARNING")
        self._msg(label, *msg)

    def debug(self, *msg):
        """
        Prints a warning
        """
        label = colors.purple("DEBUG")
        self._msg(label, *msg)

    def end(self, *msg):
        """
        Prints an end message
        """
        label = colors.purple("END")
        self._msg(label, *msg)

    def start(self, *msg):
        """
        Prints an start message
        """
        label = colors.purple("START")
        self._msg(label, *msg)

    def msg_(self, label, *msg):
        """
        Returns a message with a label
        """
        txt = self._unpack_msg(*msg)
        return "[" + label + "] " + txt

    def _msg(self, label, *msg):
        """
        Prints a message with a label
        """
        txt = self._unpack_msg(*msg)
        print("[" + label + "] " + txt)

    def _unpack_msg(self, *msg):
        """
        Convert all message elements to string
        """
        l = []
        for m in msg:
            l.append(str(m))
        return " ".join(l)



from .bokeh import Bokeh
from .altair import Altair


class Plot(Bokeh, Altair):
    """
    Class to handle charts
    """

    def __init__(self, df=None):
        """
        Initialize with an empty dataframe
        """
        self.df = df
        self.x_field = None
        self.y_field = None
        self.chart_obj = None
        self.chart_opts = dict(width=940)
        self.chart_style = None
        self.label = None
        self.engine = "bokeh"

    def chart(self, x_field=None, y_field=None, chart_type="line", opts=None, style=None, label=None):
        """
        Initialize chart options
        """
        if opts is not None:
            self.chart_opts = opts
        if style is not None:
            self.chart_style = style
        if label is not None:
            self.label = label
        self.x_field = x_field
        self.y_field = y_field
        try:
            self.chart_obj = self._get_chart(chart_type, x_field,
                                             y_field, style=style, opts=opts, label=label)
        except Exception as e:
            self.err(e, self.chart, "Can not chart data")

    def bar_(self, label=None, style=None, opts=None):
        """
        Get a bar chart
        """
        try:
            return self._get_chart("bar", style=style, opts=opts, label=label)
        except Exception as e:
            self.err(e, self.chart, "Can draw bar chart")

    def line_(self, label=None, style=None, opts=None):
        """
        Get a line chart
        """
        try:
            return self._get_chart("line", style=style, opts=opts, label=label)
        except Exception as e:
            self.err(e, self.line_, "Can draw line chart")

    def area_(self, label=None, style=None, opts=None):
        """
        Get an area chart
        """
        try:
            return self._get_chart("area", style=style, opts=opts, label=label)
        except Exception as e:
            self.err(e, self.area_, "Can draw area chart")

    def hist_(self, label=None, style=None, opts=None):
        """
        Get an historiogram chart
        """
        try:
            return self._get_chart("hist", style=style, opts=opts, label=label)
        except Exception as e:
            self.err(e, self.hist_, "Can draw historiogram")

    def errorbar_(self, label=None, style=None, opts=None):
        """
        Get a point chart
        """
        try:
            return self._get_chart("err", style=style, opts=opts, label=label)
        except Exception as e:
            self.err(e, self.errorbar_, "Can draw errorbar chart")

    def point_(self, label=None, style=None, opts=None):
        """
        Get a point chart
        """
        try:
            return self._get_chart("point", style=style, opts=opts, label=label)
        except Exception as e:
            self.err(e, self.point_, "Can draw point chart")

    def heatmap_(self, label=None, style=None, opts=None):
        """
        Get a heatmap chart
        """
        try:
            return self._get_chart("point", style=style, opts=opts, label=label)
        except Exception as e:
            self.err(e, self.heatmap_, "Can draw heatmap")

    def line_point_(self, label=None, style=None, opts=None, colors={"line": "yellow", "point": "navy"}):
        """
        Get a line and point chart
        """
        try:
            if style is None:
                style = self.chart_style
            style["color"] = colors["line"]
            l = self._get_chart("line", style=style, opts=opts, label=label)
            style["color"] = colors["point"]
            p = self._get_chart("point", style=style, opts=opts, label=label)
            return l * p
        except Exception as e:
            self.err(e, self.line_point_, "Can draw line_point chart")

    def opts(self, dictobj):
        """
        Add or update an option value to defaults
        """
        for k in dictobj:
            self.chart_opts[k] = dictobj[k]

    def style(self, dictobj):
        """
        Add or update a style value to defaults
        """
        for k in dictobj:
            self.chart_style[k] = dictobj[k]

    def _get_chart(self, chart_type, x_field=None, y_field=None, style=None, opts=None, label=None):
        """
        Get a full chart object
        """
        if x_field is None:
            if self.x_field is None:
                self.err(
                    self._get_chart, "X field is not set: please specify a parameter")
                return
            x_field = self.x_field
        if y_field is None:
            if self.y_field is None:
                self.err(
                    self._get_chart, "Y field is not set: please specify a parameter")
                return
            y_field = self.y_field
        if opts is None:
            opts = self.chart_opts
        if style is None:
            style = self.chart_style
        if x_field is None:
            x_field = self.x_field
        if y_field is None:
            y_field = self.y_field
        if self.engine == "bokeh":
            func = self._get_bokeh_chart
        elif self.engine == "altair":
            func = self._get_altair_chart
        else:
            self.err(self._get_chart, "Engine " + self.engine + " unknown")
            return
        try:
            chart = func(
                x_field, y_field, chart_type, label, opts, style)
            return chart
        except Exception as e:
            self.err(e)



import holoviews as hv


class Bokeh():
    """
    A class to handle charts with the Bokeh library
    """

    def __init__(self, df=None, db=None):
        """
        Initialize
        """
        self.df = df
        self.x_field = None
        self.y_field = None
        self.chart_obj = None
        self.chart_opts = dict(width=940)
        self.chart_style = None
        self.label = None
        self.engine = "bokeh"

    def _get_bokeh_chart(self, x_field, y_field, chart_type, label, opts, style):
        """
        Get a chart object
        """
        args = dict(data=self.df, kdims=[x_field], vdims=[y_field])
        if label is not None:
            args["label"] = label
        else:
            if self.label is not None:
                args["label"] = self.label
        chart = None
        try:
            if chart_type == "line":
                try:
                    chart = hv.Curve(**args)
                except Exception as e:
                    self.err(e)
            elif chart_type == "point":
                chart = hv.Scatter(**args)
            elif chart_type == "area":
                chart = hv.Area(**args)
            elif chart_type == "bar":
                chart = hv.Bars(**args)
            elif chart_type == "hist":
                chart = hv.Histogram(**args)
            elif chart_type == "err":
                chart = hv.ErrorBars(**args)
            endchart = chart(plot=opts, style=style)
            return endchart
        # BROKEN in Holoview 1.9.0
        # except DataError as e:
        #    msg = "Column not found in " + x_field + " and " + y_field
        #    self.err(e, msg)
        # TODO : check for column errors
        except Exception as e:
            self.err(e)
        if chart is None:
            self.err("Chart type " + chart_type +
                     " unknown", self._get_base_chart)



import pandas as pd


class Select():
    """
    Class to select data
    """

    def __init__(self, df=None):
        """
        Initialize with an empty dataframe
        """
        self.df = df

    def first_(self):
        """
        Returns the first row
        """
        return self._first()

    def _first(self):
        """
        Select the first row
        """
        try:
            val = self.df.iloc[0]
            return val
        except Exception as e:
            self.err(e, self.first, "Can not select first row")

    def limit(self, r=5):
        """
        Limit selection the a range in the main dataframe
        """
        try:
            self.df = self.df[:r]
        except Exception as e:
            self.err(e, self.limit, "Can not limit data")

    def limit_(self, r=5):
        """
        Returns a DataSwim instance with limited selection
        """
        try:
            return self.new(self.df[:r])
        except Exception as e:
            self.err(e)

    def unique_(self, column):
        """
        Returns unique values in a column
        """
        try:
            df = self.df[column].unique()
            return df
        except Exception as e:
            self.err(e, self.unique, "Can not select unique data")

    def range_(self, num, unit):
        """
        Limit the data in a time range
        """
        try:
            df = self.df[self.df.last_valid_index() -
                         pd.DateOffset(num, unit):]
            return self.clone_(df=df)
        except Exception as e:
            self.err(e, self.range_, "Can not select range data")
        return self.duplicate(df)

    def to_records_(self):
        """
        Returns a list of dictionary records from the main dataframe
        """
        try:
            dic = self.df.to_dict(orient="records")
            return dic
        except Exception as e:
            self.err(e, self.to_records_, "Can not create records")

    def nulls_(self, field):
        """
        Return all null rows
        """
        try:
            null_rows = self.df[field][self.df[field].isnull()]
        except Exception as e:
            self.err(e, self.nulls_, "Can not select null rows")
        return null_rows
